// Mini Control plug-in for VUPlayer
// Copyright (C) 2003 James Chapman
// http://www.vuplayer.com

#include <windows.h>
#include <commctrl.h>
#include <shlobj.h>
#include "resource.h"
#include "../vu_vis.h"

#define DELAY		100
#define MAXSTRING	255

char description[]="Mini Control";			// Our plug-in description
char wndclass[]="vuplayer_vis4";			// The class name for our window

BOOL CALLBACK DlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK URLProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam);
int CALLBACK BrowseProc(HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData);
void GetConfig();
void SetConfig();
void SetPalette();
void SetFont();
void FontDialog();
void ColourDialog(BOOL back);
void OpenFiles();
void OpenFolder();
void OpenURL();
void UpdateVolumeTip();
void SetTooltips();
void RemoveTooltips();
void SetSubclass();
void RemoveSubclass();

// Subclassing our control windows will allow us to redirect messages to the main DlgProc
LRESULT CALLBACK PlayProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK PauseProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK StopProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK PreviousProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK NextProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK OpenProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK VolumeProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
LRESULT CALLBACK InfoProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
WNDPROC wpPlay;
WNDPROC wpPause;
WNDPROC wpStop;
WNDPROC wpPrevious;
WNDPROC wpNext;
WNDPROC wpOpen;
WNDPROC wpVolume;
WNDPROC wpInfo;

HWND hWnd;
HWND hWndParent;
HWND hWndInfo;
HWND hWndTT;
HINSTANCE hInstance;
HICON hIcon;
HICON hIconPlay;
HICON hIconPause;
HICON hIconStop;
HICON hIconPrevious;
HICON hIconNext;
HICON hIconOpen;
HCURSOR hCursor;
HBITMAP hBitmap;
HBITMAP hOldBitmap;
HFONT hFont=0;
HFONT hOldFont;
LOGFONT logFont;
HDC hMemDC;
HDC hDC;
BYTE* pDIB;
VU_INFO* pInfo;
TOOLINFO ti;
char volumetext[16];
char strOutput[MAXSTRING+1]="";
char url[MAXSTRING+1]="";
COLORREF customcolours[16];
RECT clientrect;
RECT inforect;
RECT textrect;
RECT calcrect;
int infowidth;
int infoheight;
int elapsed;
int duration;
int x,y;
DWORD volume=0;
BOOL seek=0;
int seekpos=0;

// Our config data
DWORD xpos=0;
DWORD ypos=0;
DWORD topmost=0;
DWORD colour=0x00ffcccc;
DWORD backcolour=0x00800000;
char fontstyle[32]="ms sans serif";
char lastdir[MAX_PATH+1]="";
DWORD fontsize=8;
DWORD fontbold=0;
DWORD fontitalic=0;

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryVisual()
{
	return description;
}

HWND Init(VU_HEADER* pHdr)
{
	hWndParent=pHdr->hWndParent;			// Get a handle to the VUPlayer main window
	hInstance=pHdr->hInstance;				// Get our instance handle
	pInfo=pHdr->pInfo;						// Get a pointer to the extended info structure
	if(pInfo)
	{
		GetConfig();						// Get our config data and load the control icons
		hIcon=(HICON)LoadImage(hInstance,MAKEINTRESOURCE(IDI_VUICON),IMAGE_ICON,16,16,0);
		hIconPlay=(HICON)LoadImage(hInstance,MAKEINTRESOURCE(IDI_PLAY),IMAGE_ICON,16,16,0);
		hIconPause=(HICON)LoadImage(hInstance,MAKEINTRESOURCE(IDI_PAUSE),IMAGE_ICON,16,16,0);
		hIconStop=(HICON)LoadImage(hInstance,MAKEINTRESOURCE(IDI_STOP),IMAGE_ICON,16,16,0);
		hIconPrevious=(HICON)LoadImage(hInstance,MAKEINTRESOURCE(IDI_PREVIOUS),IMAGE_ICON,16,16,0);
		hIconNext=(HICON)LoadImage(hInstance,MAKEINTRESOURCE(IDI_NEXT),IMAGE_ICON,16,16,0);
		hIconOpen=(HICON)LoadImage(hInstance,MAKEINTRESOURCE(IDI_OPEN),IMAGE_ICON,16,16,0);
		hWnd=CreateDialog(hInstance,MAKEINTRESOURCE(IDD_DIALOG),0,DlgProc);	// Create a modeless dialog
		if(hWnd)
		{
			SetTooltips();					// Create some tooltips
			SetSubclass();					// Subclass our control windows
			hWndInfo=GetDlgItem(hWnd,ID_INFO);
			GetClientRect(hWndInfo,&clientrect);
			textrect=clientrect;
			InflateRect(&textrect,-1,-1);
			infowidth=clientrect.right-clientrect.left;
			infoheight=clientrect.bottom-clientrect.top;
			if(infowidth%8)
				infowidth+=(8-infowidth%8);
			hMemDC=CreateCompatibleDC(0);	// Create a memory device context for drawing operations
			if(hMemDC)
			{
				BYTE* buffer=new BYTE[sizeof(BITMAPINFOHEADER)+(256*4)];
				if(buffer)					// Create an 8-bit DIB (we'll only use the first 2 colours)
				{
					PBITMAPINFO pInfo=(PBITMAPINFO)buffer;
					BITMAPINFOHEADER bmpHeader;
					ZeroMemory(pInfo,sizeof(BITMAPINFO));
					ZeroMemory(&bmpHeader,sizeof(BITMAPINFOHEADER));
					bmpHeader.biCompression=BI_RGB;
					bmpHeader.biBitCount=8;
					bmpHeader.biHeight=0-infoheight;
					bmpHeader.biPlanes=1;
					bmpHeader.biSize=sizeof(BITMAPINFOHEADER);
					bmpHeader.biWidth=infowidth;
					pInfo->bmiHeader=bmpHeader;
					void* ptrDibs=0;
					hBitmap=CreateDIBSection(hMemDC,pInfo,DIB_RGB_COLORS,&ptrDibs,0,0);
					pDIB=(BYTE*)ptrDibs;
					delete [] buffer;
					if(hBitmap)
					{
						hOldBitmap=(HBITMAP)SelectObject(hMemDC,hBitmap);
						SetFont();			// Set the font for our info window
						SetPalette();		// Set the colours for our info window
						HMENU menu=GetSystemMenu(hWnd,FALSE);	// Modify the system menu
						if(menu)
						{
							InsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,0);
							InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_TOPMOST,"Always on Top");
							InsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,0);
							InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_ABOUT,"About...");
							customcolours[0]=(COLORREF)GetSysColor(COLOR_3DFACE);	// Set customcolours
							for(x=1;x<16;x++)
								customcolours[x]=customcolours[0];
						}
						SetWindowPos(hWnd,0,xpos,ypos,0,0,SWP_NOSIZE|SWP_NOZORDER);
						ShowWindow(hWnd,SW_SHOWNORMAL);
						if(topmost)
						{
							SetWindowPos(hWnd,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
							CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_CHECKED);
						}
						GetWindowRect(hWndInfo,&inforect);
						POINT pt;
						pt.x=inforect.left;
						pt.y=inforect.top;
						ScreenToClient(hWnd,&pt);
						OffsetRect(&inforect,pt.x-inforect.left,pt.y-inforect.top);
						InflateRect(&inforect,-1,-1);
						return hWnd;
					}
					SelectObject(hMemDC,hOldFont);
					DeleteObject(hFont);
					SelectObject(hMemDC,hOldBitmap);
					DeleteObject(hBitmap);
				}
				DeleteObject(hMemDC);
			}
			RemoveSubclass();
			RemoveTooltips();
			DestroyWindow(hWnd);
		}
	}
	return 0;
}

void Quit()														// Save our config data and tidy up
{
	SetConfig();
	RemoveSubclass();
	RemoveTooltips();
	SelectObject(hMemDC,hOldFont);
	DeleteObject(hFont);
	SelectObject(hMemDC,hOldBitmap);	
	DeleteObject(hBitmap);
	DeleteObject(hMemDC);
	DestroyWindow(hWnd);
}

int Render()													// Redraw the info window
{
	hDC=GetDC(hWndInfo);
	if(hDC)
	{
		ZeroMemory(pDIB,infowidth*infoheight);
		if(strlen(pInfo->lpstrTitle))							// Draw the display text
		{
			strncpy(strOutput,pInfo->lpstrArtist,MAXSTRING);
			if(strlen(strOutput))
			{
				strncat(strOutput," - ",MAXSTRING-strlen(strOutput));
				strncat(strOutput,pInfo->lpstrTitle,MAXSTRING-strlen(strOutput));
				DrawText(hMemDC,strOutput,-1,&calcrect,DT_NOPREFIX|DT_CENTER|DT_VCENTER|DT_SINGLELINE|DT_CALCRECT);
				if((calcrect.right-calcrect.left)>(textrect.right-textrect.left))
					strncpy(strOutput,pInfo->lpstrTitle,MAXSTRING);
			}
			else
				strncpy(strOutput,pInfo->lpstrTitle,MAXSTRING);
		}
		else
			ZeroMemory(strOutput,MAXSTRING);
		DrawText(hMemDC,strOutput,-1,&textrect,DT_NOPREFIX|DT_CENTER|DT_VCENTER|DT_SINGLELINE|DT_END_ELLIPSIS);
		duration=pInfo->duration;
		if(seek)
			elapsed=seekpos;
		else
			elapsed=pInfo->elapsed;
		if(elapsed&&duration)									// Draw the progress bar
		{
			if(elapsed>duration)
				elapsed=duration;
			elapsed=elapsed*(clientrect.right-clientrect.left)/duration;
			for(x=0;x<elapsed;x++)
			{	
				for(y=0;y<infoheight;y++)
					pDIB[y*infowidth+x]=!pDIB[y*infowidth+x];
			}
		}
		BitBlt(hDC,0,0,clientrect.right-clientrect.left,clientrect.bottom-clientrect.top,hMemDC,0,0,SRCCOPY);
		ReleaseDC(hWndInfo,hDC);
	}
	if(pInfo->volume!=volume)									// Check for changes in volume level
	{
		volume=pInfo->volume;
		SendMessage(GetDlgItem(hWnd,ID_VOLUME),TBM_SETPOS,1,volume);
		UpdateVolumeTip();
	}
	return DELAY;
}

/////////////////////
// Internal functions
/////////////////////

void SetPalette()				// Create a palette for the info window
{
	if(hMemDC)
	{
		RGBQUAD palette[2];
		RGBQUAD rgb;
		rgb.rgbReserved=0;
		rgb.rgbBlue=(BYTE)(backcolour>>16&0xff);
		rgb.rgbGreen=(BYTE)(backcolour>>8&0xff);
		rgb.rgbRed=(BYTE)(backcolour&0xff);
		palette[0]=rgb;
		rgb.rgbBlue=(BYTE)(colour>>16&0xff);
		rgb.rgbGreen=(BYTE)(colour>>8&0xff);
		rgb.rgbRed=(BYTE)(colour&0xff);
		palette[1]=rgb;
		for(int index=0;index<10;index++)
		{
			if(SetDIBColorTable(hMemDC,0,2,palette))
				break;
			Sleep(10);
		}
		SetTextColor(hMemDC,colour);
		SetBkColor(hMemDC,backcolour);
		SetBkMode(hMemDC,TRANSPARENT);
	}
}

void SetFont()					// Set the font for the info window
{
	if(hMemDC)
	{
		if(hFont)
		{
			SelectObject(hMemDC,hOldFont);
			DeleteObject(hFont);
		}
		ZeroMemory(&logFont,sizeof(LOGFONT));
		logFont.lfHeight=-MulDiv(fontsize,GetDeviceCaps(hMemDC,LOGPIXELSY),72);
		if(fontitalic)
			logFont.lfItalic=1;
		if(fontbold)
			logFont.lfWeight=FW_BOLD;
		strncpy(logFont.lfFaceName,fontstyle,31);
		hFont=CreateFontIndirect(&logFont);
		hOldFont=(HFONT)SelectObject(hMemDC,hFont);
	}
}

void UpdateVolumeTip()			// Update the text for the volume slider tooltip
{
	strcpy(volumetext,"Volume: ");
	char buffer[4];
	itoa(volume,buffer,10);
	strcat(volumetext,buffer);
	strcat(volumetext,"%");
	SendMessage(hWndTT,TTM_UPDATETIPTEXT,0,(LPARAM)&ti);
}

void FontDialog()				// Display a font common dialog box
{
	CHOOSEFONT choose;
	ZeroMemory(&choose,sizeof(CHOOSEFONT));
	choose.hwndOwner=hWnd;
	choose.lpLogFont=&logFont;
	choose.nSizeMin=6;
	choose.nSizeMax=12;
	choose.Flags=CF_LIMITSIZE|CF_SCREENFONTS|CF_INITTOLOGFONTSTRUCT;
	choose.lStructSize=sizeof(CHOOSEFONT);
	if(ChooseFont(&choose))
	{
		strncpy(fontstyle,logFont.lfFaceName,31);
		fontsize=choose.iPointSize/10;
		if(choose.nFontType&BOLD_FONTTYPE)
			fontbold=1;
		else
			fontbold=0;
		if(choose.nFontType&ITALIC_FONTTYPE)
			fontitalic=1;
		else
			fontitalic=0;
		SetFont();
	}
}

void ColourDialog(BOOL back)	// Display a color common dialog box
{
	CHOOSECOLOR choose;
	ZeroMemory(&choose,sizeof(CHOOSECOLOR));
	choose.lStructSize=sizeof(CHOOSECOLOR);
	choose.Flags=CC_ANYCOLOR|CC_FULLOPEN|CC_RGBINIT;
	choose.lpCustColors=customcolours;
	if(back)
		choose.rgbResult=backcolour;
	else
		choose.rgbResult=colour;
	if(ChooseColor(&choose))
	{
		if(back)
			backcolour=choose.rgbResult;
		else
			colour=choose.rgbResult;
		SetPalette();
	}
}

void OpenFiles()				// Display an Open common dialog box
{
	char* buffer=new char[32768];
	if(!buffer)
		return;
	ZeroMemory(buffer,32768);
	OPENFILENAME openFile;
	ZeroMemory(&openFile,sizeof(OPENFILENAME));
	openFile.lStructSize=sizeof(OPENFILENAME);
	openFile.hwndOwner=hWnd;
	openFile.lpstrFilter="Supported music files\0*.mod;*.s3m;*.xm;*.it;*.mp2;*.mp3;*.ogg;*.flac;*.asf;*.wma;*.wav;*.vpl;*.m3u;*.pls\0All files (*.*)\0*.*\0\0";
	openFile.lpstrFile=buffer;
	openFile.nMaxFile=32768;
	openFile.lpstrTitle="Open File(s)";
	openFile.Flags=OFN_ALLOWMULTISELECT|OFN_EXPLORER|OFN_HIDEREADONLY;
	if(GetOpenFileName(&openFile))
	{
		char dirname[MAX_PATH]="";
		strncpy(dirname,buffer,MAX_PATH);
		if((strlen(dirname)>0)&&(!openFile.nFileExtension))				// Multiple files selected
		{
			char filename[MAX_PATH]="";
			char* output=dirname+strlen(dirname)-1;
			strncpy(filename,output,1);
			if(strcmp(filename,"\\"))
				strcat(dirname,"\\");
			output=openFile.lpstrFile+openFile.nFileOffset;
			BOOL add=0;
			while(*output)
			{
				strncpy(filename,dirname,MAX_PATH);
				strncat(filename,output,MAX_PATH-strlen(filename));
				if(add)
					SendMessage(hWndParent,VU_MESSAGE,VU_ADDFILE,(LPARAM)filename);
				else
				{
					if(SendMessage(hWndParent,VU_MESSAGE,VU_PLAYFILE,(LPARAM)filename))
						add=1;
					else
						add=0;
				}
				output+=(strlen(output)+1);
			}
		}
		else
			SendMessage(hWndParent,VU_MESSAGE,VU_PLAYFILE,(LPARAM)dirname);
	}
	delete [] buffer;
}

void OpenFolder()				// Display a Browse for Folder dialog
{
	BROWSEINFO browseInfo;
	ZeroMemory(&browseInfo,sizeof(BROWSEINFO));
	browseInfo.hwndOwner=hWnd;
	browseInfo.pszDisplayName=lastdir;
	browseInfo.lpszTitle="Choose a folder to scan for music:";
	browseInfo.lpfn=BrowseProc;
	browseInfo.lParam=(LPARAM)lastdir;
	LPITEMIDLIST idList=SHBrowseForFolder(&browseInfo);
	if(idList&&SHGetPathFromIDList(idList,lastdir))
		SendMessage(hWndParent,VU_MESSAGE,VU_PLAYFILE,(LPARAM)lastdir);
	LPMALLOC pMalloc;
	if(SHGetMalloc(&pMalloc)==NOERROR)
	{
		pMalloc->Free(idList);
		pMalloc->Release();
	}
}

void OpenURL()					// Display an Open URL dialog
{
	if(DialogBox(hInstance,MAKEINTRESOURCE(IDD_OPENURL),hWnd,URLProc))
		SendMessage(hWndParent,VU_MESSAGE,VU_PLAYFILE,(LPARAM)url);
}

void GetConfig()				// Get our config data from the registry
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer\\Mini Control",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[MAX_PATH];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"xpos",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			xpos=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"ypos",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			ypos=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"colour",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			colour=*(COLORREF*)buffer;
		if(RegQueryValueEx(hKey,"backcolour",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			backcolour=*(COLORREF*)buffer;
		if(RegQueryValueEx(hKey,"topmost",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			topmost=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"fontsize",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			fontsize=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"fontbold",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			fontbold=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"fontitalic",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			fontitalic=*(DWORD*)buffer;
		bufSize=32;
		if(RegQueryValueEx(hKey,"fontstyle",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			strncpy(fontstyle,(char*)buffer,31);
		bufSize=MAX_PATH;
		if(RegQueryValueEx(hKey,"lastdir",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			strncpy(lastdir,(char*)buffer,MAX_PATH);
		RegCloseKey(hKey);
	}
}

void SetConfig()				// Save our config data to the registry
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer\\Mini Control",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"xpos",0,REG_DWORD,(BYTE*)&xpos,4);
		RegSetValueEx(hKey,"ypos",0,REG_DWORD,(BYTE*)&ypos,4);
		RegSetValueEx(hKey,"topmost",0,REG_DWORD,(BYTE*)&topmost,4);
		RegSetValueEx(hKey,"colour",0,REG_DWORD,(BYTE*)&colour,4);
		RegSetValueEx(hKey,"backcolour",0,REG_DWORD,(BYTE*)&backcolour,4);
		RegSetValueEx(hKey,"fontstyle",0,REG_SZ,(BYTE*)fontstyle,strlen(fontstyle)+1);
		RegSetValueEx(hKey,"fontsize",0,REG_DWORD,(BYTE*)&fontsize,4);
		RegSetValueEx(hKey,"fontbold",0,REG_DWORD,(BYTE*)&fontbold,4);
		RegSetValueEx(hKey,"fontitalic",0,REG_DWORD,(BYTE*)&fontitalic,4);
		RegSetValueEx(hKey,"lastdir",0,REG_SZ,(BYTE*)lastdir,strlen(lastdir)+1);
		RegCloseKey(hKey);
	}
}

void SetTooltips()				// Create tooltips for the controls
{
	hWndTT=CreateWindow(TOOLTIPS_CLASS,0,0,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,CW_USEDEFAULT,hWnd,0,hInstance,0);
	if(hWndTT)
	{
		ZeroMemory(&ti,sizeof(TOOLINFO));
		ti.cbSize=sizeof(TOOLINFO);
		ti.uFlags=TTF_SUBCLASS|TTF_IDISHWND;
		ti.hinst=hInstance;
		ti.hwnd=hWnd;
		ti.uId=(UINT)(GetDlgItem(hWnd,IDB_PLAY));
		ti.lpszText="Play";
		SendMessage(hWndTT,TTM_ADDTOOL,0,(LPARAM)&ti);
		ti.uId=(UINT)(GetDlgItem(hWnd,IDB_PAUSE));
		ti.lpszText="Pause";
		SendMessage(hWndTT,TTM_ADDTOOL,0,(LPARAM)&ti);
		ti.uId=(UINT)(GetDlgItem(hWnd,IDB_STOP));
		ti.lpszText="Stop";
		SendMessage(hWndTT,TTM_ADDTOOL,0,(LPARAM)&ti);
		ti.uId=(UINT)(GetDlgItem(hWnd,IDB_PREVIOUS));
		ti.lpszText="Previous Track";
		SendMessage(hWndTT,TTM_ADDTOOL,0,(LPARAM)&ti);
		ti.uId=(UINT)(GetDlgItem(hWnd,IDB_NEXT));
		ti.lpszText="Next Track";
		SendMessage(hWndTT,TTM_ADDTOOL,0,(LPARAM)&ti);
		ti.uId=(UINT)(GetDlgItem(hWnd,IDB_OPEN));
		ti.lpszText="Open File(s)";
		SendMessage(hWndTT,TTM_ADDTOOL,0,(LPARAM)&ti);
		ti.uId=(UINT)(GetDlgItem(hWnd,ID_VOLUME));
		ti.lpszText=volumetext;
		SendMessage(hWndTT,TTM_ADDTOOL,0,(LPARAM)&ti);
		UpdateVolumeTip();
	}
}

void RemoveTooltips()
{
	DestroyWindow(hWndTT);
}

void SetSubclass()				// Subclass the control windows, storing all the old window procedures 
{
	wpPlay=(WNDPROC)SetWindowLong(GetDlgItem(hWnd,IDB_PLAY),GWL_WNDPROC,(long)PlayProc);
	wpPause=(WNDPROC)SetWindowLong(GetDlgItem(hWnd,IDB_PAUSE),GWL_WNDPROC,(long)PauseProc);
	wpStop=(WNDPROC)SetWindowLong(GetDlgItem(hWnd,IDB_STOP),GWL_WNDPROC,(long)StopProc);
	wpPrevious=(WNDPROC)SetWindowLong(GetDlgItem(hWnd,IDB_PREVIOUS),GWL_WNDPROC,(long)PreviousProc);
	wpNext=(WNDPROC)SetWindowLong(GetDlgItem(hWnd,IDB_NEXT),GWL_WNDPROC,(long)NextProc);
	wpOpen=(WNDPROC)SetWindowLong(GetDlgItem(hWnd,IDB_OPEN),GWL_WNDPROC,(long)OpenProc);
	wpVolume=(WNDPROC)SetWindowLong(GetDlgItem(hWnd,ID_VOLUME),GWL_WNDPROC,(long)VolumeProc);
	wpInfo=(WNDPROC)SetWindowLong(GetDlgItem(hWnd,ID_INFO),GWL_WNDPROC,(long)InfoProc);
}

void RemoveSubclass()			// Reset the window procedures for the controls
{
	SetWindowLong(GetDlgItem(hWnd,IDB_PLAY),GWL_WNDPROC,(long)wpPlay);
	SetWindowLong(GetDlgItem(hWnd,IDB_PAUSE),GWL_WNDPROC,(long)wpPause);
	SetWindowLong(GetDlgItem(hWnd,IDB_STOP),GWL_WNDPROC,(long)wpStop);
	SetWindowLong(GetDlgItem(hWnd,IDB_PREVIOUS),GWL_WNDPROC,(long)wpPrevious);
	SetWindowLong(GetDlgItem(hWnd,IDB_NEXT),GWL_WNDPROC,(long)wpNext);
	SetWindowLong(GetDlgItem(hWnd,IDB_OPEN),GWL_WNDPROC,(long)wpOpen);
	SetWindowLong(GetDlgItem(hWnd,ID_VOLUME),GWL_WNDPROC,(long)wpVolume);
	SetWindowLong(GetDlgItem(hWnd,ID_INFO),GWL_WNDPROC,(long)wpInfo);
}

// Our subclass window procedures, courtesy of cut & paste 
LRESULT CALLBACK PlayProc(HWND hwndControl, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		PostMessage(hWnd,message,wParam,lParam);						// Redirect to parent window
		return 0;
	}
	return CallWindowProc(wpPlay,hwndControl,message,wParam,lParam);	// Use default handler
}

LRESULT CALLBACK PauseProc(HWND hwndControl, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		PostMessage(hWnd,message,wParam,lParam);
		return 0;
	}
	return CallWindowProc(wpPause,hwndControl,message,wParam,lParam);
}

LRESULT CALLBACK StopProc(HWND hwndControl, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		PostMessage(hWnd,message,wParam,lParam);
		return 0;
	}
	return CallWindowProc(wpStop,hwndControl,message,wParam,lParam);
}

LRESULT CALLBACK PreviousProc(HWND hwndControl, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		PostMessage(hWnd,message,wParam,lParam);
		return 0;
	}
	return CallWindowProc(wpPrevious,hwndControl,message,wParam,lParam);
}

LRESULT CALLBACK NextProc(HWND hwndControl, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		PostMessage(hWnd,message,wParam,lParam);
		return 0;
	}
	return CallWindowProc(wpNext,hwndControl,message,wParam,lParam);
}

LRESULT CALLBACK OpenProc(HWND hwndControl, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		PostMessage(hWnd,message,wParam,lParam);
		return 0;
	case WM_RBUTTONDOWN:
		SetCapture(hwndControl);
		return 0;
	case WM_RBUTTONUP:
		if(GetCapture()==hwndControl)
		{
			ReleaseCapture();
			POINT pt;
			RECT rect;
			GetClientRect(hwndControl,&rect);
			pt.x=LOWORD(lParam);
			pt.y=HIWORD(lParam);
			if(PtInRect(&rect,pt))
			{
				GetCursorPos(&pt);
				HMENU menu=CreatePopupMenu();
				InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_OPENURL,"Open URL...");
				InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_OPENFOLDER,"Open Folder...");
				InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_OPENFILES,"Open File(s)...");
				TrackPopupMenu(menu,TPM_RIGHTBUTTON,pt.x,pt.y,0,hWnd,0);
			}
		}
		return 0;
	}
	return CallWindowProc(wpOpen,hwndControl,message,wParam,lParam);
}

LRESULT CALLBACK VolumeProc(HWND hwndControl, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		PostMessage(hWnd,message,wParam,lParam);
		return 0;
	case WM_RBUTTONDOWN:
		SendMessage(hWndParent,VU_MESSAGE,VU_SETVOLUME,50);
		return 0;
	}
	return CallWindowProc(wpVolume,hwndControl,message,wParam,lParam);
}

LRESULT CALLBACK InfoProc(HWND hwndControl, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_KEYDOWN:
	case WM_SYSKEYDOWN:
		PostMessage(hWnd,message,wParam,lParam);
		return 0;
	case WM_MOUSEMOVE:
		if(seek)
		{
			short pos=LOWORD(lParam);
			RECT rect;
			GetClientRect(hwndControl,&rect);
			if(pos<rect.left)
				pos=(short)rect.left;
			else
			{
				if(pos>rect.right)
					pos=(short)rect.right;
			}
			seekpos=(pos-rect.left)*pInfo->duration/(rect.right-rect.left);
			Render();
		}
		return 0;
	case WM_LBUTTONDOWN:
		{
			POINT pt;
			RECT rect;
			GetClientRect(hwndControl,&rect);
			pt.x=LOWORD(lParam);
			pt.y=HIWORD(lParam);
			if(PtInRect(&rect,pt)&&(pInfo->duration))
			{
				seek=1;
				seekpos=(pt.x-rect.left)*pInfo->duration/(rect.right-rect.left);
				SetCapture(hwndControl);
			}
		}
		return 0;
	case WM_LBUTTONUP:
		if(GetCapture()==hwndControl)
		{
			ReleaseCapture();
			SendMessage(hWndParent,VU_MESSAGE,VU_SETPLAYBACKPOS,seekpos);
			seek=0;
			seekpos=0;
		}
		return 0;
	case WM_RBUTTONDOWN:
		SetCapture(hwndControl);
		return 0;
	case WM_RBUTTONUP:
		if(GetCapture()==hwndControl)
		{
			ReleaseCapture();
			POINT pt;
			RECT rect;
			GetClientRect(hwndControl,&rect);
			pt.x=LOWORD(lParam);
			pt.y=HIWORD(lParam);
			if(PtInRect(&rect,pt))
			{
				GetCursorPos(&pt);
				HMENU menu=CreatePopupMenu();
				InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_BACKCOLOUR,"Background Color...");
				InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_COLOUR,"Font Color...");
				InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_FONTSTYLE,"Font Style...");
				TrackPopupMenu(menu,TPM_RIGHTBUTTON,pt.x,pt.y,0,hWnd,0);
			}
		}
		return 0;
	}
	return CallWindowProc(wpInfo,hwndControl,message,wParam,lParam);
}

int CALLBACK BrowseProc(HWND hwnd, UINT uMsg, LPARAM lParam, LPARAM lpData)		// Browse callback proc
{
	if(uMsg==BFFM_INITIALIZED)													// Set the initial dir
		SendMessage(hwnd,BFFM_SETSELECTION,TRUE,lpData);
	return 0;
}

BOOL CALLBACK URLProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)	// URL dialog WindowProc
{
	switch(uMsg)
	{
	case WM_CLOSE:
		PostMessage(hwndDlg,WM_COMMAND,IDCANCEL,0);
		return 1;
	case WM_COMMAND:
		switch(LOWORD(wParam))
        {
			case IDOK:
				EndDialog(hwndDlg,1);
				GetDlgItemText(hwndDlg,ID_URL,url,MAXSTRING);
				return 1;
			case IDCANCEL:
				EndDialog(hwndDlg,0);
				return 1;
		}
	}
	return 0;
}

BOOL CALLBACK DlgProc(HWND hwndDlg, UINT uMsg, WPARAM wParam, LPARAM lParam)	// Main DialogProc
{
	switch(uMsg)
	{
	case WM_INITDIALOG:										// Initialize the various controls
		SendMessage(hwndDlg,WM_SETICON,ICON_SMALL,(LPARAM)hIcon);
		SendMessage(GetDlgItem(hwndDlg,IDB_PLAY),BM_SETIMAGE,IMAGE_ICON,(LPARAM)hIconPlay);
		SendMessage(GetDlgItem(hwndDlg,IDB_PAUSE),BM_SETIMAGE,IMAGE_ICON,(LPARAM)hIconPause);
		SendMessage(GetDlgItem(hwndDlg,IDB_STOP),BM_SETIMAGE,IMAGE_ICON,(LPARAM)hIconStop);
		SendMessage(GetDlgItem(hwndDlg,IDB_PREVIOUS),BM_SETIMAGE,IMAGE_ICON,(LPARAM)hIconPrevious);
		SendMessage(GetDlgItem(hwndDlg,IDB_NEXT),BM_SETIMAGE,IMAGE_ICON,(LPARAM)hIconNext);
		SendMessage(GetDlgItem(hwndDlg,IDB_OPEN),BM_SETIMAGE,IMAGE_ICON,(LPARAM)hIconOpen);
		SendMessage(GetDlgItem(hwndDlg,ID_VOLUME),TBM_SETRANGEMIN,0,0);
		SendMessage(GetDlgItem(hwndDlg,ID_VOLUME),TBM_SETRANGEMAX,0,100);
		SendMessage(GetDlgItem(hwndDlg,ID_VOLUME),TBM_SETTICFREQ,10,0);
		SendMessage(GetDlgItem(hwndDlg,ID_VOLUME),TBM_SETPAGESIZE,0,10);
		SendMessage(GetDlgItem(hwndDlg,ID_VOLUME),TBM_SETPOS,1,pInfo->volume);
		break;
	case WM_CLOSE:
		PostQuitMessage(0);
		break;
	case WM_KEYDOWN:										// Pass KEYDOWN messages to VUPlayer
		PostMessage(hWndParent,uMsg,wParam,lParam);
		break;
	case WM_SYSKEYDOWN:										// Special case to deal with F10
		if((int)wParam==VK_F10)
		{
			PostMessage(hWndParent,WM_KEYDOWN,(WPARAM)VK_F10,0);
			return 1;
		}
		break;
	case WM_PAINT:											// Repaint our window
		Render();
		ValidateRect(hwndDlg,&inforect);
		break;
	case WM_MOVING:											// Keep track of our window position
		{
			LPRECT pRect=(LPRECT)lParam;
			xpos=pRect->left;
			ypos=pRect->top;
		}
		break;
	case WM_HSCROLL:										// Keep track of the volume position
		{
			DWORD pos=SendMessage(GetDlgItem(hwndDlg,ID_VOLUME),TBM_GETPOS,0,0);
			if(pos!=pInfo->volume)
			{
				SendMessage(hWndParent,VU_MESSAGE,VU_SETVOLUME,pos);
				volume=pInfo->volume;
				UpdateVolumeTip();
			}
		}
		break;
	case WM_SYSCOMMAND:
		if(wParam==ID_ABOUT)								// Display a basic About box
			MessageBox(hwndDlg,"Mini Control plug-in for VUPlayer  \nCopyright  2003 James Chapman  \nwww.vuplayer.com","About Mini Control",MB_OK|MB_ICONINFORMATION);
		if(wParam==ID_TOPMOST)								// Enable/disable topmost status
		{
			HMENU menu=GetSystemMenu(hwndDlg,FALSE);
			if(menu)
			{
				if(topmost)
				{
					SetWindowPos(hwndDlg,HWND_NOTOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
					CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_UNCHECKED);
					topmost=0;
				}
				else
				{
					SetWindowPos(hwndDlg,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
					CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_CHECKED);
					topmost=1;
				}
			}
		}
		break;
	case WM_COMMAND:
		switch(LOWORD(wParam))
		{
		case ID_FONTSTYLE:
			FontDialog();
			break;
		case ID_COLOUR:
			ColourDialog(0);
			break;
		case ID_BACKCOLOUR:
			ColourDialog(1);
			break;
		case ID_OPENFILES:
			OpenFiles();
			break;
		case ID_OPENFOLDER:
			OpenFolder();
			break;
		case ID_OPENURL:
			OpenURL();
			break;
		case IDB_PLAY:
			if(HIWORD(wParam)==BN_CLICKED)
				SendMessage(hWndParent,VU_MESSAGE,VU_PLAY,0);
			break;
		case IDB_PAUSE:
			if(HIWORD(wParam)==BN_CLICKED)
				SendMessage(hWndParent,VU_MESSAGE,VU_PAUSE,0);
			break;
		case IDB_STOP:
			if(HIWORD(wParam)==BN_CLICKED)
			{
				SendMessage(hWndParent,VU_MESSAGE,VU_STOP,0);
				Render();
			}
			break;
		case IDB_PREVIOUS:
			if(HIWORD(wParam)==BN_CLICKED)
			{
				SendMessage(hWndParent,VU_MESSAGE,VU_PREVIOUSTRACK,0);
				Render();
			}
			break;
		case IDB_NEXT:
			if(HIWORD(wParam)==BN_CLICKED)
			{
				SendMessage(hWndParent,VU_MESSAGE,VU_NEXTTRACK,0);
				Render();
			}
			break;
		case IDB_OPEN:
			if(HIWORD(wParam)==BN_CLICKED)
				OpenFiles();
			break;
		}
		break;
	case WM_DROPFILES:										// Handle dropped files
		{
			BOOL add=0;
			if(GetKeyState(VK_CONTROL)&0x8000)
				add=1;
			HDROP hDropInfo=(HDROP)wParam;
			char buffer[MAX_PATH]="";
			UINT filecount=DragQueryFile(hDropInfo,0xffffffff,buffer,0);
			for(UINT index=0;index<filecount;index++)
			{
				DragQueryFile(hDropInfo,index,buffer,MAX_PATH);
				if(add)
					SendMessage(hWndParent,VU_MESSAGE,VU_ADDFILE,(LPARAM)buffer);
				else
				{
					if(SendMessage(hWndParent,VU_MESSAGE,VU_PLAYFILE,(LPARAM)buffer))
						add=1;
					else
						add=0;
				}
			}
		}
		break;
	}
	return 0;
}
