// Spectrum Analyser visual plug-in for VUPlayer
// Copyright (C) 2003 James Chapman
// http://www.vuplayer.com

#include <windows.h>
#include <math.h>
#include "resource.h"
#include "../vu_vis.h"

#define MINWIDTH	240
#define MINHEIGHT	180
#define LINEARSCALE	4096
#define DELAY		20

char description[]="Spectrum Analyser";		// Our plug-in description
char wndclass[]="vuplayer_vis2";			// The class name for our window

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
void SetPalette();
void ColourDialog(BOOL peakcolour);
void GetConfig();
void SetConfig();

HWND hWnd;
HWND hWndParent;
HINSTANCE hInstance;
HICON hIcon;
HCURSOR hCursor;
HBITMAP hBitmap;
HBITMAP hOldBitmap;
HDC hMemDC;
HDC hDC;
BYTE* pDIB;
DWORD dwFFT;
float* pFFT;
WORD* linearFFT;
double dValue;
COLORREF customcolours[16];
DWORD maxwidth=GetSystemMetrics(SM_CXSCREEN);
DWORD maxheight=GetSystemMetrics(SM_CYSCREEN);

// Our config data
DWORD xpos=0;
DWORD ypos=0;
DWORD width=320;
DWORD height=240;
DWORD maximised=0;
DWORD topmost=0;
COLORREF colour1=0x0000ffff;
COLORREF colour2=0x000000ff;

// Temporary variables used during rendering
DWORD clientwidth;
DWORD clientheight;
DWORD scopewidth;
DWORD scopeheight;
RECT clientrect;
DWORD index;
long count;
long point;
long centre;
BYTE pointcolour;
WORD wValue;
int fftpos;

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryVisual()
{
	return description;
}

HWND Init(VU_HEADER* pHdr)
{
	pFFT=pHdr->pFFT;				// Get a pointer to the FFT data
	dwFFT=pHdr->FFTsize;			// Make a note of the array size
	hInstance=pHdr->hInstance;		// Get our instance handle
	hWndParent=pHdr->hWndParent;	// Get a handle to the VUPlayer main window
	linearFFT=new WORD[dwFFT];		// Set up our linear FFT array
	if(!linearFFT)
		return 0;
	ZeroMemory(linearFFT,dwFFT*2);
	GetConfig();					// Get our config data (screen position & colour)
	hMemDC=CreateCompatibleDC(0);	// Create a memory device context for drawing operations
	if(hMemDC)						// Create a top-down, 8-bit (256 colour) DIB section
	{
		BYTE* buffer=new BYTE[sizeof(BITMAPINFOHEADER)+(256*4)];
		if(buffer)
		{
			PBITMAPINFO pInfo=(PBITMAPINFO)buffer;
			BITMAPINFOHEADER bmpHeader;
			ZeroMemory(pInfo,sizeof(BITMAPINFO));
			ZeroMemory(&bmpHeader,sizeof(BITMAPINFOHEADER));
			bmpHeader.biCompression=BI_RGB;
			bmpHeader.biBitCount=8;
			bmpHeader.biHeight=0-maxheight;
			bmpHeader.biPlanes=1;
			bmpHeader.biSize=sizeof(BITMAPINFOHEADER);
			bmpHeader.biWidth=maxwidth;
			pInfo->bmiHeader=bmpHeader;
			void* ptrDibs=0;
			hBitmap=CreateDIBSection(hMemDC,pInfo,DIB_RGB_COLORS,&ptrDibs,0,0);
			pDIB=(BYTE*)ptrDibs;
			delete [] buffer;
			if(hBitmap)							// Select the DIB into our memDC and set the palette
			{
				hOldBitmap=(HBITMAP)SelectObject(hMemDC,hBitmap);
				SetPalette();
				WNDCLASS wc;
				ZeroMemory(&wc,sizeof(WNDCLASS));
				wc.lpfnWndProc=WndProc;
				wc.hInstance=pHdr->hInstance;
				wc.lpszClassName=wndclass;
				hIcon=LoadIcon(hInstance,MAKEINTRESOURCE(IDI_VUICON));
				hCursor=LoadCursor(hInstance,MAKEINTRESOURCE(IDC_POINTER));
				wc.hIcon=hIcon;
				wc.hCursor=hCursor;
				if(RegisterClass(&wc))			// Register a class and create a window for our vis
				{
					hWnd=CreateWindowEx(WS_EX_ACCEPTFILES,wndclass,description,WS_POPUPWINDOW|WS_CAPTION|WS_MINIMIZEBOX|WS_MAXIMIZEBOX|WS_THICKFRAME,xpos,ypos,width,height,0,0,pHdr->hInstance,0);
					if(hWnd)
					{
						HMENU menu=GetSystemMenu(hWnd,FALSE);	// Modify the system menu
						if(menu)
						{
							InsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,0);
							InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_TOPMOST,"Always on Top");
							InsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,0);
							InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_COLOUR1,"Base Color...");
							InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_COLOUR2,"Peak Color...");
							InsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,0);
							InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_ABOUT,"About...");
							customcolours[0]=(COLORREF)GetSysColor(COLOR_3DFACE);	// Set customcolours
							for(index=1;index<16;index++)
								customcolours[index]=customcolours[0];
						}
						if(maximised)
							ShowWindow(hWnd,SW_SHOWMAXIMIZED);
						else
							ShowWindow(hWnd,SW_SHOWNORMAL);
						if(topmost)
						{
							SetWindowPos(hWnd,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
							CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_CHECKED);
						}
						return hWnd;
					}
					UnregisterClass(wndclass,hInstance);
				}
				SelectObject(hMemDC,hOldBitmap);
				DeleteObject(hBitmap);
			}
		}
		DeleteObject(hMemDC);
	}
	return 0;
}

void Quit()														// Save our config data and tidy up
{
	delete [] linearFFT;
	SetConfig();						
	SelectObject(hMemDC,hOldBitmap);	
	DeleteObject(hBitmap);
	DeleteObject(hMemDC);
	DestroyWindow(hWnd);
	UnregisterClass(wndclass,hInstance);
}

int Render()
{
	if(GetClientRect(hWnd,&clientrect))
	{
		clientwidth=clientrect.right-clientrect.left;			// Get the client dimensions
		clientheight=clientrect.bottom-clientrect.top;
		if(!clientheight)
			return DELAY;
		if(clientwidth>maxwidth)
			scopewidth=maxwidth;
		else
			scopewidth=clientwidth;
		if(clientheight>maxheight)
			scopeheight=maxheight;
		else
			scopeheight=clientheight;
		centre=scopeheight/2;									// Work out the centre line position
		if(!(scopeheight%2))
			centre--;
		ZeroMemory(pDIB,maxwidth*scopeheight);					// Clear the DIB section
		index=0;
		while(index<dwFFT)										// Convert FFT data to a linear scale
		{
			if(pFFT[index]<0.0001)
				wValue=0;
			else
			{
				dValue=log10(pFFT[index]);
				wValue=(WORD)((4.0+dValue)*(LINEARSCALE/4));
				if(wValue>LINEARSCALE)
					wValue=LINEARSCALE;
			}
			if(wValue>=linearFFT[index])
				linearFFT[index]=(linearFFT[index]+wValue)/2;
			else
			{
				if(linearFFT[index]>(LINEARSCALE/64))
					linearFFT[index]-=LINEARSCALE/64;
				else
					linearFFT[index]=0;
			}
			index++;
		}
		index=1;
		while(index<(scopewidth-3))
		{
			fftpos=(index-1)*dwFFT/scopewidth;					// Determine position in FFT array
			if(fftpos%2)										// Plot the left channel 
				fftpos--;
			count=linearFFT[fftpos]*centre/LINEARSCALE;
			while(count>0)
			{
				point=(centre-count)*maxwidth+index;
				pointcolour=(BYTE)(255*(count-1)/centre+1);
				pDIB[point++]=pointcolour;
				pDIB[point++]=pointcolour;
				pDIB[point]=pointcolour;
				count--;
			}
			count=linearFFT[fftpos+1]*centre/LINEARSCALE;		// Plot the right channel
			while(count>0)
			{
				point=(centre+count)*maxwidth+index;
				pointcolour=(BYTE)(255*(count-1)/centre+1);
				pDIB[point++]=pointcolour;
				pDIB[point++]=pointcolour;
				pDIB[point]=pointcolour;
				count--;
			}
			index+=4;
		}
		hDC=GetDC(hWnd);
		if(hDC)													// Blit the image to the screen
		{
			if(clientwidth>maxwidth||clientheight>maxheight)	// Deal with increased screen resolution
			{
				if(clientwidth>maxwidth)
				{
					scopewidth=(clientwidth-maxwidth)/2;
					BitBlt(hDC,0,0,scopewidth,clientheight,hMemDC,0,0,BLACKNESS);
					BitBlt(hDC,scopewidth+maxwidth,0,clientwidth-scopewidth-maxwidth,clientheight,hMemDC,0,0,BLACKNESS);
				}
				else
					scopewidth=0;
				if(clientheight>maxheight)
				{
					scopeheight=(clientheight-maxheight)/2;
					BitBlt(hDC,0,0,clientwidth,scopeheight,hMemDC,0,0,BLACKNESS);
					BitBlt(hDC,0,scopeheight+maxheight,clientwidth,clientheight-scopeheight-maxheight,hMemDC,0,0,BLACKNESS);
					clientheight=maxheight;
				}
				else
					scopeheight=0;
				if(clientwidth>maxwidth)
					clientwidth=maxwidth;
				BitBlt(hDC,scopewidth,scopeheight,clientwidth,clientheight,hMemDC,0,0,SRCCOPY);
			}
			else
				BitBlt(hDC,0,0,clientwidth,clientheight,hMemDC,0,0,SRCCOPY);
			ReleaseDC(hWnd,hDC);
		}
	}
	return DELAY;												// Do it all again real soon
}

/////////////////////
// Internal functions
/////////////////////

void SetPalette()												// Create a nice gradient for our DIB
{
	if(hMemDC)
	{
		RGBQUAD palette[256];
		ZeroMemory(&palette,256*sizeof(RGBQUAD));
		RGBQUAD rgb;
		rgb.rgbReserved=0;
		int startred=colour1&0xff;
		int startgreen=colour1>>8&0xff;
		int startblue=colour1>>16&0xff;
		int endred=colour2&0xff;
		int endgreen=colour2>>8&0xff;
		int endblue=colour2>>16&0xff;
		for(int index=1;index<=255;index++)
		{
			rgb.rgbBlue=(BYTE)(startblue+index*(endblue-startblue)/255);
			rgb.rgbGreen=(BYTE)(startgreen+index*(endgreen-startgreen)/255);
			rgb.rgbRed=(BYTE)(startred+index*(endred-startred)/255);
			palette[index]=rgb;
		}
		for(endblue=0;endblue<10;endblue++)
		{
			if(SetDIBColorTable(hMemDC,0,256,palette))		// Update palette (retry if necessary)
				break;
			Sleep(10);
		}
	}
}

void ColourDialog(BOOL peakcolour)							// Display a color common dialog box
{
	CHOOSECOLOR choose;
	ZeroMemory(&choose,sizeof(CHOOSECOLOR));
	choose.lStructSize=sizeof(CHOOSECOLOR);
	choose.Flags=CC_ANYCOLOR|CC_FULLOPEN|CC_RGBINIT;
	choose.lpCustColors=customcolours;
	if(peakcolour)
		choose.rgbResult=colour2;
	else
		choose.rgbResult=colour1;
	if(ChooseColor(&choose))
	{
		if(peakcolour)
			colour2=choose.rgbResult;
		else
			colour1=choose.rgbResult;
		SetPalette();
	}
}

void GetConfig()											// Get our config data from the registry
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer\\Spectrum Analyser",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[4];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"xpos",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			xpos=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"ypos",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			ypos=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"width",0,0,buffer,&bufSize)==ERROR_SUCCESS)
		{
			width=*(DWORD*)buffer;
			if((width>maxwidth)||(width<MINWIDTH))
				width=320;
		}
		if(RegQueryValueEx(hKey,"height",0,0,buffer,&bufSize)==ERROR_SUCCESS)
		{
			height=*(DWORD*)buffer;
			if((height>maxheight)||(height<MINHEIGHT))
				height=240;
		}
		if(RegQueryValueEx(hKey,"maximised",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			maximised=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"topmost",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			topmost=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"colour1",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			colour1=*(COLORREF*)buffer;
		if(RegQueryValueEx(hKey,"colour2",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			colour2=*(COLORREF*)buffer;
		RegCloseKey(hKey);
	}
}

void SetConfig()											// Save our config data to the registry
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer\\Spectrum Analyser",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"xpos",0,REG_DWORD,(BYTE*)&xpos,4);
		RegSetValueEx(hKey,"ypos",0,REG_DWORD,(BYTE*)&ypos,4);
		RegSetValueEx(hKey,"width",0,REG_DWORD,(BYTE*)&width,4);
		RegSetValueEx(hKey,"height",0,REG_DWORD,(BYTE*)&height,4);
		RegSetValueEx(hKey,"maximised",0,REG_DWORD,(BYTE*)&maximised,4);
		RegSetValueEx(hKey,"topmost",0,REG_DWORD,(BYTE*)&topmost,4);
		RegSetValueEx(hKey,"colour1",0,REG_DWORD,(BYTE*)&colour1,4);
		RegSetValueEx(hKey,"colour2",0,REG_DWORD,(BYTE*)&colour2,4);
		RegCloseKey(hKey);
	}
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_SYSCOMMAND:
		if(wParam==ID_ABOUT)								// Display a basic About box
		{
			MessageBox(hwnd,"Spectrum Analyser plug-in for VUPlayer  \nCopyright  2003 James Chapman  \nwww.vuplayer.com","About Spectrum Analyser",MB_OK|MB_ICONINFORMATION);
			return 0;
		}
		if(wParam==ID_COLOUR1)
		{
			ColourDialog(0);
			return 0;
		}
		if(wParam==ID_COLOUR2)
		{
			ColourDialog(1);
			return 0;
		}
		if(wParam==ID_TOPMOST)								// Enable/disable topmost status
		{
			HMENU menu=GetSystemMenu(hWnd,FALSE);
			if(menu)
			{
				if(topmost)
				{
					SetWindowPos(hWnd,HWND_NOTOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
					CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_UNCHECKED);
					topmost=0;
				}
				else
				{
					SetWindowPos(hWnd,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
					CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_CHECKED);
					topmost=1;
				}
			}
			return 0;
		}
		break;
	case WM_SIZING:											// Keep track of our window size
		{
			LPRECT pRect=(LPRECT)lParam;
			if((pRect->right-pRect->left)<MINWIDTH)
			{
				if((wParam==WMSZ_RIGHT)||(wParam==WMSZ_BOTTOMRIGHT)||(wParam==WMSZ_TOPRIGHT))
					pRect->right=pRect->left+MINWIDTH;
				else
					pRect->left=pRect->right-MINWIDTH;
			}
			if((pRect->bottom-pRect->top)<MINHEIGHT)
			{
				if((wParam==WMSZ_BOTTOM)||(wParam==WMSZ_BOTTOMLEFT)||(wParam==WMSZ_BOTTOMRIGHT))
					pRect->bottom=pRect->top+MINHEIGHT;
				else
					pRect->top=pRect->bottom-MINHEIGHT;
			}
			xpos=pRect->left;
			ypos=pRect->top;
			width=pRect->right-pRect->left;
			height=pRect->bottom-pRect->top;
		}
		return 1;
	case WM_MOVING:											// Keep track of our window position
		{
			LPRECT pRect=(LPRECT)lParam;
			xpos=pRect->left;
			ypos=pRect->top;
		}
		return 1;
	case WM_SIZE:											// Keep track of our maximised status
		if(IsWindowVisible(hwnd))
		{
			if(wParam==SIZE_MAXIMIZED)
				maximised=1;
			else
				maximised=0;
		}
		return 0;
	case WM_ERASEBKGND:										// We don't want to erase the background
		return 1;
	case WM_PAINT:											// Repaint our window
		{
			PAINTSTRUCT ps;
			HDC dc=BeginPaint(hwnd,&ps);
			GetClientRect(hwnd,&clientrect);
			BitBlt(dc,0,0,clientrect.right-clientrect.left,clientrect.bottom-clientrect.top,hMemDC,0,0,SRCCOPY);
			EndPaint(hwnd,&ps);
		}
        return 0;
	case WM_CREATE:
		return 0;
	case WM_CLOSE:											// Signal to VUPlayer that we want to quit
		PostQuitMessage(0);
		return 0;
	case WM_KEYDOWN:										// Pass KEYDOWN messages to VUPlayer
		PostMessage(hWndParent,message,wParam,lParam);
		return 0;
	case WM_SYSKEYDOWN:										// Special case to deal with F10
		if(((int)wParam==VK_F10)&&(!(lParam>>29&1)))
		{
			PostMessage(hWndParent,WM_KEYDOWN,(WPARAM)VK_F10,lParam);
			return 0;
		}
		break;
	case WM_DROPFILES:										// Handle dropped files
		{
			BOOL add=0;
			if(GetKeyState(VK_CONTROL)&0x8000)
				add=1;
			HDROP hDropInfo=(HDROP)wParam;
			char buffer[MAX_PATH]="";
			UINT filecount=DragQueryFile(hDropInfo,0xffffffff,buffer,0);
			for(UINT index=0;index<filecount;index++)
			{
				DragQueryFile(hDropInfo,index,buffer,MAX_PATH);
				if(add)
					SendMessage(hWndParent,VU_MESSAGE,VU_ADDFILE,(LPARAM)buffer);
				else
				{
					if(SendMessage(hWndParent,VU_MESSAGE,VU_PLAYFILE,(LPARAM)buffer))
						add=1;
					else
						add=0;
				}
			}
		}
		return 0;
	}
	return DefWindowProc(hwnd,message,wParam,lParam);
}
