// VU Meter visual plug-in for VUPlayer
// Copyright (C) 2003 James Chapman
// http://www.vuplayer.com

#include <windows.h>
#include "resource.h"
#include "../vu_vis.h"
#include "vu_stuff.h"

#define DELAY		20

char description[]="VU Meter";				// Our plug-in description
char wndclass[]="vuplayer_vis3";			// The class name for our window

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam);
void GetConfig();
void SetConfig();
void UnpackMask();
void SetBackground();
void DrawPin(BOOL left);

HWND hWnd;
HWND hWndParent;
HINSTANCE hInstance;
HICON hIcon;
HCURSOR hCursor;
HBITMAP hOldBitmap;
HDC hMemDC;
HDC hDC;
HBITMAP hLeftBitmap;
HBITMAP hRightBitmap;
BYTE* pLeftDIB;
BYTE* pRightDIB;
BYTE VUMask[VU_WIDTH*VU_HEIGHT];
BYTE VULeft[VU_WIDTH*VU_HEIGHT*3];
BYTE VURight[VU_WIDTH*VU_HEIGHT*3];
COLORREF customcolours[16];

// Our config data
DWORD xpos=0;
DWORD ypos=0;
DWORD topmost=0;
COLORREF colour=0x00ffaaaa;

// Temporary variables used during rendering
DWORD* pLevel;
BYTE* pDIB;
BYTE* pBack;
WORD* previous;
WORD previousLeft=0;
WORD previousRight=0;
WORD volume;
DWORD index;
DWORD pos;
BYTE value;

/////////////////////
// Exported functions
/////////////////////

LPCTSTR QueryVisual()
{
	return description;
}

HWND Init(VU_HEADER* pHdr)
{
	hWndParent=pHdr->hWndParent;			// Get a handle to the VUPlayer main window
	hInstance=pHdr->hInstance;				// Get our instance handle
	pLevel=pHdr->pLevel;					// Get a pointer to the volume level
	GetConfig();							// Get our config data
	hMemDC=CreateCompatibleDC(0);			// Create a memory device context for drawing operations
	if(hMemDC)								// Create a pair of 24-bit bottom-up DIB sections
	{
		BITMAPINFO bmpInfo;
		BITMAPINFOHEADER bmpHeader;
		ZeroMemory(&bmpHeader,sizeof(BITMAPINFOHEADER));
		bmpHeader.biBitCount=24;
		bmpHeader.biCompression=BI_RGB;
		bmpHeader.biHeight=VU_HEIGHT;
		bmpHeader.biPlanes=1;
		bmpHeader.biSize=sizeof(BITMAPINFOHEADER);
		bmpHeader.biWidth=VU_WIDTH;
		bmpInfo.bmiHeader=bmpHeader;
		void* ptrDibs=0;
		hLeftBitmap=CreateDIBSection(hMemDC,&bmpInfo,DIB_RGB_COLORS,&ptrDibs,0,0);
		pLeftDIB=(BYTE*)ptrDibs;
		hRightBitmap=CreateDIBSection(hMemDC,&bmpInfo,DIB_RGB_COLORS,&ptrDibs,0,0);
		pRightDIB=(BYTE*)ptrDibs;
		if(hLeftBitmap&&hRightBitmap)
		{
			UnpackMask();					// Unpack the VUMeter mask
			SetBackground();				// Set the background colour
			WNDCLASS wc;
			ZeroMemory(&wc,sizeof(WNDCLASS));
			wc.lpfnWndProc=WndProc;
			wc.hInstance=pHdr->hInstance;
			wc.lpszClassName=wndclass;
			hIcon=LoadIcon(hInstance,MAKEINTRESOURCE(IDI_VUICON));
			hCursor=LoadCursor(hInstance,MAKEINTRESOURCE(IDC_POINTER));
			wc.hIcon=hIcon;
			wc.hCursor=hCursor;
			if(RegisterClass(&wc))			// Register a class and create a fixed size vis window
			{
				hWnd=CreateWindowEx(WS_EX_ACCEPTFILES,wndclass,description,WS_POPUPWINDOW|WS_CAPTION|WS_MINIMIZEBOX,xpos,ypos,VU_WIDTH*2,VU_HEIGHT,0,0,pHdr->hInstance,0);
				if(hWnd)
				{
					RECT clientrect;						// Adjust the client area
					RECT windowrect;
					GetClientRect(hWnd,&clientrect);
					GetWindowRect(hWnd,&windowrect);
					MoveWindow(hWnd,windowrect.left,windowrect.top,(windowrect.right-windowrect.left)+(VU_WIDTH*2-(clientrect.right-clientrect.left)),(windowrect.bottom-windowrect.top)+(VU_HEIGHT-(clientrect.bottom-clientrect.top)),FALSE);
					HMENU menu=GetSystemMenu(hWnd,FALSE);	// Modify the system menu
					if(menu)
					{
						InsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,0);
						InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_TOPMOST,"Always on Top");
						InsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,0);
						InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_COLOUR,"Select Color...");
						InsertMenu(menu,0,MF_BYPOSITION|MF_SEPARATOR,0,0);
						InsertMenu(menu,0,MF_BYPOSITION|MF_STRING,ID_ABOUT,"About...");
						customcolours[0]=(COLORREF)GetSysColor(COLOR_3DFACE);	// Set customcolours
						for(index=1;index<16;index++)
							customcolours[index]=customcolours[0];
					}
					ShowWindow(hWnd,SW_SHOWNORMAL);
					if(topmost)
					{
						SetWindowPos(hWnd,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
						CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_CHECKED);
					}
					return hWnd;
				}
				UnregisterClass(wndclass,hInstance);
			}
			DeleteObject(hLeftBitmap);
			DeleteObject(hRightBitmap);
		}
		DeleteObject(hMemDC);
	}
	return 0;
}

void Quit()													// Save our config data and tidy up
{
	SetConfig();						
	DeleteObject(hLeftBitmap);
	DeleteObject(hRightBitmap);
	DeleteObject(hMemDC);
	DestroyWindow(hWnd);
	UnregisterClass(wndclass,hInstance);
}

int Render()												// Do some damping when we set the volume
{
	volume=LOWORD(*pLevel)>>8;
	if(volume>=previousLeft)
		volume=previousLeft+(volume-previousLeft)/2+(volume-previousLeft)%2;
	else
	{
		if(previousLeft>64)
			volume=previousLeft-2;
		else
		{
			if(previousLeft)
				volume=previousLeft-1;
			else
				volume=0;
		}
	}
	if(volume>128)
		volume=128;
	DrawPin(1);												// Draw the left meter
	volume=HIWORD(*pLevel)>>8;
	if(volume>=previousRight)
		volume=previousRight+(volume-previousRight)/2+(volume-previousRight)%2;
	else
	{
		if(previousRight>64)
			volume=previousRight-2;
		else
		{
			if(previousRight)
				volume=previousRight-1;
			else
				volume=0;
		}
	}
	if(volume>128)
		volume=128;
	DrawPin(0);												// Draw the right meter
	hDC=GetDC(hWnd);
	if(hDC)													// Blit the images to the screen
	{
		hOldBitmap=(HBITMAP)SelectObject(hMemDC,hLeftBitmap);
		BitBlt(hDC,0,0,VU_WIDTH,VU_HEIGHT,hMemDC,0,0,SRCCOPY);
		SelectObject(hMemDC,hOldBitmap);
		hOldBitmap=(HBITMAP)SelectObject(hMemDC,hRightBitmap);
		BitBlt(hDC,VU_WIDTH,0,VU_WIDTH,VU_HEIGHT,hMemDC,0,0,SRCCOPY);
		SelectObject(hMemDC,hOldBitmap);
		ReleaseDC(hWnd,hDC);
	}
	return DELAY;
}

/////////////////////
// Internal functions
/////////////////////

void UnpackMask()
{
	BYTE testbyte=0;
	BYTE bytecount=0;
	DWORD index=0;
	BYTE count=0;
	DWORD outindex=0;
	while(index<VU_PACKED)
	{
		switch(blankVU[index])
		{
		case 0:
			bytecount=blankVU[++index];
			count=0;
			while(count<bytecount)
			{
				VUMask[outindex++]=0;
				count++;
			}
			break;
		case 0xff:
			bytecount=blankVU[++index];
			count=0;
			while(count<bytecount)
			{
				VUMask[outindex++]=0xff;
				count++;
			}
			break;
		default:
			VUMask[outindex++]=blankVU[index];
		}
		index++;
	}
}

void SetBackground()
{
	DWORD blue=(colour>>16)&0xff;
	DWORD green=(colour>>8)&0xff;
	DWORD red=colour&0xff;
	DWORD index=0;
	DWORD value=0;
	while(index<(VU_HEIGHT*VU_WIDTH))		// Set the background colour (flip for the right meter)
	{
		VULeft[index*3]=(BYTE)(blue+baseVU[index]*(255-blue)/255);
		VULeft[index*3+1]=(BYTE)(green+baseVU[index]*(255-green)/255);
		VULeft[index*3+2]=(BYTE)(red+baseVU[index]*(255-red)/255);
		VURight[((index/VU_WIDTH+1)*VU_WIDTH-(index%VU_WIDTH+1))*3]=VULeft[index*3];
		VURight[((index/VU_WIDTH+1)*VU_WIDTH-(index%VU_WIDTH+1))*3+1]=VULeft[index*3+1];
		VURight[((index/VU_WIDTH+1)*VU_WIDTH-(index%VU_WIDTH+1))*3+2]=VULeft[index*3+2];
		index++;
	}
	index=0;
	while(index<(VU_HEIGHT*VU_WIDTH))		// Overlay the background mask
	{
		value=VUMask[index];
		if(value)
		{
			if(value==0xff)
			{
				VULeft[index*3]=0;
				VULeft[index*3+1]=0;
				VULeft[index*3+2]=0;
				VURight[index*3]=0;
				VURight[index*3+1]=0;
				VURight[index*3+2]=0;
			}
			else
			{
				VULeft[index*3]=(BYTE)(VULeft[index*3]-((value*VULeft[index*3])>>8));
				VULeft[index*3+1]=(BYTE)(VULeft[index*3+1]-((value*VULeft[index*3+1])>>8));
				VULeft[index*3+2]=(BYTE)(VULeft[index*3+2]-((value*VULeft[index*3+2])>>8));
				VURight[index*3]=(BYTE)(VURight[index*3]-((value*VURight[index*3])>>8));
				VURight[index*3+1]=(BYTE)(VURight[index*3+1]-((value*VURight[index*3+1])>>8));
				VURight[index*3+2]=(BYTE)(VURight[index*3+2]-((value*VURight[index*3+2])>>8));
			}
		}
		index++;
	}
	index=0;
	while(index<(VU_HEIGHT*VU_WIDTH))		// Update the DIB sections
	{
		*(pLeftDIB+index*3)=VULeft[index*3];
		*(pLeftDIB+index*3+1)=VULeft[index*3+1];
		*(pLeftDIB+index*3+2)=VULeft[index*3+2];
		*(pRightDIB+index*3)=VURight[index*3];
		*(pRightDIB+index*3+1)=VURight[index*3+1];
		*(pRightDIB+index*3+2)=VURight[index*3+2];
		index++;
	}
}

void DrawPin(BOOL left)
{
	if(left)
	{
		previous=&previousLeft;
		pBack=VULeft;
		pDIB=pLeftDIB;
	}
	else
	{
		previous=&previousRight;
		pBack=VURight;
		pDIB=pRightDIB;
	}
	index=0;
	while(index<VUPinSizes[*previous])		// Blank out the previous pin position
	{
		pos=VUpins[*previous][index];
		pos&=0xffff;
		*(pDIB+pos*3)=pBack[pos*3];
		*(pDIB+pos*3+1)=pBack[pos*3+1];
		*(pDIB+pos*3+2)=pBack[pos*3+2];
		index++;
	}
	index=0;
	while(index<VUPinSizes[volume])			// Draw the new pin position
	{
		pos=VUpins[volume][index];
		value=(BYTE)(pos>>16);
		pos&=0xffff;
		if(value==0xff)						// Black out the point
		{
			*(pDIB+pos*3)=0;
			*(pDIB+pos*3+1)=0;
			*(pDIB+pos*3+2)=0;
		}
		else								// Merge value with background
		{
			*(pDIB+pos*3)=pBack[pos*3]-((value*pBack[pos*3])>>8);
			*(pDIB+pos*3+1)=pBack[pos*3+1]-((value*pBack[pos*3+1])>>8);
			*(pDIB+pos*3+2)=pBack[pos*3+2]-((value*pBack[pos*3+2])>>8);
		}
		index++;
	}
	*previous=volume;
}

void GetConfig()											// Get our config data from the registry
{
	HKEY hKey;
	if(RegOpenKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer\\VUMeter",0,KEY_QUERY_VALUE,&hKey)==ERROR_SUCCESS)
	{
		BYTE buffer[4];
		DWORD bufSize=4;
		if(RegQueryValueEx(hKey,"xpos",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			xpos=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"ypos",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			ypos=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"topmost",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			topmost=*(DWORD*)buffer;
		if(RegQueryValueEx(hKey,"colour",0,0,buffer,&bufSize)==ERROR_SUCCESS)
			colour=*(COLORREF*)buffer;
		RegCloseKey(hKey);
	}
}

void SetConfig()											// Save our config data to the registry
{
	DWORD disp;
	HKEY hKey;
	if(RegCreateKeyEx(HKEY_CURRENT_USER,"Software\\VUPlayer\\VUMeter",0,0,REG_OPTION_NON_VOLATILE,KEY_ALL_ACCESS,0,&hKey,&disp)==ERROR_SUCCESS)
	{
		RegSetValueEx(hKey,"xpos",0,REG_DWORD,(BYTE*)&xpos,4);
		RegSetValueEx(hKey,"ypos",0,REG_DWORD,(BYTE*)&ypos,4);
		RegSetValueEx(hKey,"topmost",0,REG_DWORD,(BYTE*)&topmost,4);
		RegSetValueEx(hKey,"colour",0,REG_DWORD,(BYTE*)&colour,4);
		RegCloseKey(hKey);
	}
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	switch(message)
	{
	case WM_SYSCOMMAND:
		if(wParam==ID_ABOUT)								// Display a basic About box
		{
			MessageBox(hwnd,"VU Meter plug-in for VUPlayer  \nCopyright  2003 James Chapman  \nwww.vuplayer.com","About VU Meter",MB_OK|MB_ICONINFORMATION);
			return 0;
		}
		if(wParam==ID_COLOUR)								// Display a color common dialog box
		{
			CHOOSECOLOR choose;
			ZeroMemory(&choose,sizeof(CHOOSECOLOR));
			choose.lStructSize=sizeof(CHOOSECOLOR);
			choose.Flags=CC_ANYCOLOR|CC_FULLOPEN|CC_RGBINIT;
			choose.lpCustColors=customcolours;
			choose.rgbResult=colour;
			if(ChooseColor(&choose))
			{
				colour=choose.rgbResult;
				SetBackground();
			}
			return 0;
		}
		if(wParam==ID_TOPMOST)								// Enable/disable topmost status
		{
			HMENU menu=GetSystemMenu(hWnd,FALSE);
			if(menu)
			{
				if(topmost)
				{
					SetWindowPos(hWnd,HWND_NOTOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
					CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_UNCHECKED);
					topmost=0;
				}
				else
				{
					SetWindowPos(hWnd,HWND_TOPMOST,0,0,0,0,SWP_NOMOVE|SWP_NOSIZE);
					CheckMenuItem(menu,ID_TOPMOST,MF_BYCOMMAND|MF_CHECKED);
					topmost=1;
				}
			}
			return 0;
		}
		break;
	case WM_MOVING:											// Keep track of our window position
		{
			LPRECT pRect=(LPRECT)lParam;
			xpos=pRect->left;
			ypos=pRect->top;
		}
		return 1;
	case WM_ERASEBKGND:										// We don't want to erase the background
		return 1;
	case WM_PAINT:											// Repaint our window
		{
			PAINTSTRUCT ps;
			HDC dc=BeginPaint(hwnd,&ps);
			hOldBitmap=(HBITMAP)SelectObject(hMemDC,hLeftBitmap);
			BitBlt(dc,0,0,VU_WIDTH,VU_HEIGHT,hMemDC,0,0,SRCCOPY);
			SelectObject(hMemDC,hOldBitmap);
			hOldBitmap=(HBITMAP)SelectObject(hMemDC,hRightBitmap);
			BitBlt(dc,VU_WIDTH,0,VU_WIDTH,VU_HEIGHT,hMemDC,0,0,SRCCOPY);
			SelectObject(hMemDC,hOldBitmap);
			EndPaint(hwnd,&ps);
		}
        return 0;
	case WM_CREATE:
		return 0;
	case WM_CLOSE:											// Signal to VUPlayer that we want to quit
		PostQuitMessage(0);
		return 0;
	case WM_KEYDOWN:										// Pass KEYDOWN messages to VUPlayer
		PostMessage(hWndParent,message,wParam,lParam);
		return 0;
	case WM_SYSKEYDOWN:										// Special case to deal with F10
		if(((int)wParam==VK_F10)&&(!(lParam>>29&1)))
		{
			PostMessage(hWndParent,WM_KEYDOWN,(WPARAM)VK_F10,lParam);
			return 0;
		}
		break;
	case WM_DROPFILES:										// Handle dropped files
		{
			BOOL add=0;
			if(GetKeyState(VK_CONTROL)&0x8000)
				add=1;
			HDROP hDropInfo=(HDROP)wParam;
			char buffer[MAX_PATH]="";
			UINT filecount=DragQueryFile(hDropInfo,0xffffffff,buffer,0);
			for(UINT index=0;index<filecount;index++)
			{
				DragQueryFile(hDropInfo,index,buffer,MAX_PATH);
				if(add)
					SendMessage(hWndParent,VU_MESSAGE,VU_ADDFILE,(LPARAM)buffer);
				else
				{
					if(SendMessage(hWndParent,VU_MESSAGE,VU_PLAYFILE,(LPARAM)buffer))
						add=1;
					else
						add=0;
				}
			}
		}
		return 0;
	}
	return DefWindowProc(hwnd,message,wParam,lParam);
}

