// VUPlayer visual plug-in header file

#define VU_MESSAGE	34466		//  Use this message ID to send commands to the VUPlayer main window
//		WPARAM						LPARAM					RETURN VALUE
#define VU_PLAY				1	//	-						-
#define VU_PAUSE			2	//	-						-
#define VU_STOP				3	//	-						-
#define VU_PREVIOUSTRACK	4	//	-						-
#define VU_NEXTTRACK		5	//	-						-
#define VU_SKIPBACKWARD		6	//	-						-
#define VU_SKIPFORWARD		7	//	-						-
#define VU_VOLUMEDOWN		8	//	-						-
#define VU_VOLUMEUP			9	//	-						-
#define VU_SETSTOPATEND		10	//	0:off,1:on				-
#define VU_SETVOLUMEFADE	11	//	0:off,1:on				-
#define VU_SETVOLUMEMUTE	12	//	0:off,1:on				-
#define VU_SETVOLUME		13	//	(int)volume	(0 to 100)	-
#define VU_SETRANDOM		14	//	0:off,1:on				-
#define VU_SETREPEAT		15	//	0:off,1:on				-
#define VU_SETREPEATLIST	16	//	0:off,1:on				-
#define VU_SETPLAYBACKPOS	17	//	(int)seconds			0:fail,1:success
#define VU_PLAYFILE			18	//	(LPCTSTR)filename		0:fail,1:success (playback started)
#define VU_ADDFILE			19	//	(LPCTSTR)filename		0:fail,1:success (>0 files added)

// Supported filetypes for the VU_INFO structure
#define FILETYPE_MOD		1
#define FILETYPE_MOD6		2
#define FILETYPE_MOD8		3
#define FILETYPE_S3M		4
#define FILETYPE_XM			5
#define FILETYPE_IT			6
#define FILETYPE_MP3		7
#define FILETYPE_WAV		8
#define FILETYPE_OGG		9
#define FILETYPE_WMA		10
#define FILETYPE_FLAC		11
#define FILETYPE_URL		12
#define FILETYPE_MP4		13
#define FILETYPE_APE		14
#define FILETYPE_WV			15
#define FILETYPE_MTM		16
#define FILETYPE_CDA		31

// Supported state flags for the VU_INFO structure
#define STATE_PLAYING		1
#define STATE_PAUSED		2
#define STATE_MUTE			4
#define STATE_STOPATEND		8
#define STATE_VOLUMEFADE	16
#define STATE_RANDOM		32
#define STATE_REPEATTRACK	64
#define STATE_REPEATLIST	128

typedef struct
{
	LPCTSTR lpstrArtist;	// Information about the currently loaded file
	LPCTSTR lpstrTitle;		//
	LPCTSTR lpstrAlbum;		//
	LPCTSTR lpstrGenre;		//
	LPCTSTR lpstrFilepath;	//
	WORD year;				//
	WORD filetype;			// Filetype (see above)
	WORD bitrate;			// [kbps]
	WORD channels;			//
	DWORD freq;				// [Hz]
	DWORD filesize;			// [bytes]
	DWORD duration;			// [seconds]
	DWORD elapsed;			// Current elapsed time [seconds]
	DWORD volume;			// Current main volume level [0 to 100]
	DWORD state;			// State flags (see above)
	DWORD reserved;			// Reserved for future use
} VU_INFO;

typedef struct{
	HWND hWndParent;		// Handle to the VUPlayer main window
	HINSTANCE hInstance;	// DLL instance handle
	DWORD datasize;			// Size of sample data array			[8192 as of VUPlayer v1.9]
	DWORD FFTsize;			// Size (in floats) of FFT data array	[1024 as of VUPlayer v1.9]
	DWORD* pLevel;			// Volume level [left=LOWORD, right=HIWORD, ranging from 0 to 32768]
	DWORD* pRate;			// The current sample rate in Hz
	BYTE* pData;			// Stereo (LLRRLLRR...) sample data		[16-bit signed]
	float* pFFT;			// Stereo (LRLRLRLR...) FFT data		[value range is 0 to 1.0f]
	VU_INFO* pInfo;			// Extended information (see above)
} VU_HEADER;

// VUPlayer will scan all files in the visuals folder on startup and attempt to call this function.
// Return the description of your plug-in to be used in the VUPlayer visuals menu.
LPCTSTR QueryVisual();

// Do all your setup here and return the window handle to your vis on success, 0 on failure.
HWND Init(VU_HEADER* pHdr);

// Put all your cleanup code here.
void Quit();

// Return the desired millisecond delay until the next call.
// If there is some problem, return a negative value to have VUPlayer call Quit().
int Render();

